#This hold many function used in the HOD fitter
from __future__ import print_function

import sys
import numpy as np
import collections
import os.path as path


try:
   import pandas as pd
   pd_flag=1
except:
   pd_flag=0
   print('''pandas not found: you can install pandas from
            https://pypi.python.org/pypi/pandas/0.18.0/#downloads
            The code will work without pandas but pandas will speed up loading large text files
            ''')

#This uses pandas if available otherwise numpy loadtxt to read files
def load_txt(fname):
   if(pd_flag==1):
      tb=pd.read_table(fname,delim_whitespace=True,comment='#',header=None,
              dtype=np.float)
      tb=tb.as_matrix()
   else:
      tb=np.loadtxt(fname)
   return tb


def load_halos(SHOD):
    #This function loads the halos catalogs from different halod types
    #It stores the following essential information of the halos
    #Halos positions, mass and velocities with velocity dispersion
    #it should also store some randomly selected particels if available.

    #everything will held in a ordered dictionary
    SHOD['halos']=collections.OrderedDict()

    if(SHOD['halo_filetype']=='RunPBtxt'):
        htmp=load_txt(SHOD['halo_cat'])
    
        #apply any selections if needed
        ind=htmp[:,3]>SHOD['Mcut'] #only positive non-zero halo mass or somem minimum from config file

        #now assign important propwerites
        SHOD['halos']['xyz']  = htmp[ind,:3]*SHOD['Lbox']
        SHOD['halos']['xyztype']='nonunit' #to say whether halo co-ordinates are in unit box or not
        SHOD['halos']['vxyz'] = htmp[ind,4:7]
        SHOD['halos']['Mvir']= htmp[ind,3]
        SHOD['halos']['Mhalo']= htmp[ind,3]
        SHOD['halos']['sigv'] = htmp[ind,7]

    if(SHOD['halo_filetype'] in ['RunPBRockstar','LegacyRockstar']):
        for tkey in ['xyz','vxyz','Mvir','sigv','r200','rs','rsr200']:
            SHOD['halos'][tkey]=np.array([])
        #first load the halo catalog
        Kmap={'xyz':[8,11],'vxyz':[11,14],'Mvir':[27], 'sigv':[7],'r200':[4],'rs':[44]}

        SHOD['halos']['xyztype']='nonunit' #to say whether halo co-ordinates are in unit box or not
        colmap={'xx':8,'yy':9,'zz':10}
        fnum=0
        #hfname='halos_1.0000.%d.ascii'%(fnum)
        hfname='%s.%d.ascii'%(SHOD['halo_cat'],fnum)
        print(hfname)
        while(path.isfile(SHOD['halo_dir']+hfname)):
            print(fnum,hfname)
            htmp=load_txt(SHOD['halo_dir']+hfname)
            #apply any selections if needed
            ind=htmp[:,Kmap['Mvir'][0]]>SHOD['Mcut'] #only positive non-zero halo mass
            for tkey in Kmap.keys():
               if(len(Kmap[tkey])==1):
                   SHOD['halos'][tkey] = accumulate_arrays(SHOD['halos'][tkey],htmp[ind,Kmap[tkey][0]])
               else:
                   SHOD['halos'][tkey] = accumulate_arrays(SHOD['halos'][tkey],
                            htmp[ind,Kmap[tkey][0]:Kmap[tkey][1]])
            #htmp.close()
            fnum=fnum+1
            hfname='%s.%d.ascii'%(SHOD['halo_cat'],fnum)

        #now assign important propwerites
        SHOD['halos']['Mhalo']= SHOD['halos']['Mvir']
        SHOD['halos']['rsr200'] = SHOD['halos']['rs']/SHOD['halos']['r200']

    if(SHOD['halo_filetype'] in ['MDPL2','MDPL1']):
        #first load the halo catalog from FITS file
        import fitsio as F
        fin=F.FITS(SHOD['halo_cat'])

        #apply any selections if needed
        #only positive non-zero halo mass and vrms
        sel_str='Mvir>%f && vrms>0'%(SHOD['Mcut'])
        #whether to use subhalo or not by default its included
        if(SHOD['subhalo']==0):
           sel_str=sel_str+' && pid==-1'
           print('---- Removing subhalos from the file by pid==-1')

        #Apply the selection
        ind=fin[1].where(sel_str)

        #now assign important propwerites
        SHOD['halos']['xyz']  = np.column_stack([fin[1]['xx'][ind],fin[1]['yy'][ind],fin[1]['zz'][ind]])
        SHOD['halos']['xyztype']='nonunit' #to say whether halo co-ordinates are in unit box or not
        SHOD['halos']['vxyz']  = np.column_stack([fin[1]['vx'][ind],fin[1]['vy'][ind],fin[1]['vz'][ind]])
        SHOD['halos']['Mvir'] = fin[1]['Mvir'][ind]
        SHOD['halos']['Mhalo']= SHOD['halos']['Mvir']
        SHOD['halos']['sigv'] = fin[1]['vrms'][ind] 
        SHOD['halos']['r200'] = fin[1]['rvir'][ind]
        rs=fin[1]['rs'][ind]
        SHOD['halos']['rsr200'] = rs/SHOD['halos']['r200']

        fin.close()

    if(SHOD['halo_filetype']=='OUTERRIM'): #read outer rim with genericio
        sys.path.insert(0,'/home/salam/source/genericio/python/')
        import genericio
        #get the halo mass and apply the cuts
        Halo_mass_arr=genericio.gio_read(SHOD['halo_cat'],'fof_halo_mass')
        #apply any selections if needed
        #only positive non-zero halo mass 
        ind=np.where(Halo_mass_arr>SHOD['Mcut'])

        #Assign the Mvir 
        SHOD['halos']['Mvir'] = 2.0*Halo_mass_arr[ind] 
        # The above factor of 2 is needed to match the mass velocity dispersion relation and closer halo mass function
        #get x,y,z co-ordinates and velocity
        for pp,prop in enumerate(['x','y','z']):
           this_co=genericio.gio_read(SHOD['halo_cat'],'fof_halo_center_%s'%prop)
           this_ve=genericio.gio_read(SHOD['halo_cat'],'fof_halo_mean_v%s'%prop)
           if(pp==0):
              SHOD['halos']['xyz']=this_co[ind]
              SHOD['halos']['vxyz']=this_ve[ind]
           else:
              SHOD['halos']['xyz']=np.column_stack([SHOD['halos']['xyz'],this_co[ind]])
              SHOD['halos']['vxyz']=np.column_stack([SHOD['halos']['vxyz'],this_ve[ind]])

        #free some memory
        this_co=None; this_ve=None; Halo_mass_arr=None
        SHOD['halos']['xyztype']='nonunit' #to say whether halo co-ordinates are in unit box or not
        #get the velocity dispersion
        vrms_arr=genericio.gio_read(SHOD['halo_cat'],'fof_halo_vel_disp')
        SHOD['halos']['sigv'] = vrms_arr[ind]; vrms_arr=None
        #get the r200
        SHOD['halos']['r200'] = HVD.get_r200_Outerrim(Mvir_halo=SHOD['halos']['Mvir'])
        #first interpolate the concentration velocity dispersion relations
        #This assumes spherically symmetric NFW halos to evaluate this analytically
        print('Mhalo_min max:',SHOD['halos']['Mvir'].min(),SHOD['halos']['Mvir'].max())
        print('vrms lims:',SHOD['halos']['sigv'].min(),SHOD['halos']['sigv'].max())
        LMvir_min=np.log10(SHOD['halos']['Mvir'].min())
        LMvir_max=np.log10(SHOD['halos']['Mvir'].max())
        Idic_conc=HVD.setup_concentration_vrms_interpolate(LMvir_min=LMvir_min,LMvir_max=LMvir_max,
                conc_min=1.0,conc_max=100,vrms_min=SHOD['halos']['sigv'].min(),vrms_max=SHOD['halos']['sigv'].max(),plots=0)
        #get the inverse of concentration by using interpolated relation with scatter same as mean
        SHOD['halos']['rsr200'] =1.0/HVD.evaluate_conc_vrms_interp(Idic_conc,SHOD['halos']['Mvir'],SHOD['halos']['sigv'])
        SHOD['halos']['Mhalo']=SHOD['halos']['Mvir']


    if(SHOD['halo_filetype']=='bolshoitxt'):
        #first load the halo catalog
        htmp=load_txt(SHOD['halo_cat'])

        #apply any selections if needed
        ind=htmp[:,10]>SHOD['Mcut'] #only positive non-zero halo mass

        #now assign important propwerites
        SHOD['halos']['xyz']  = htmp[ind,17:20]
        SHOD['halos']['xyztype']='nonunit' #to say whether halo co-ordinates are in unit box or not
        SHOD['halos']['vxyz'] = htmp[ind,20:23]
        SHOD['halos']['Mvir']= htmp[ind,10]
        SHOD['halos']['Mhalo']= SHOD['halos']['Mvir']
        SHOD['halos']['sigv'] = htmp[ind,13] 
        SHOD['halos']['r200'] = htmp[ind,11]
        rs=htmp[ind,12]
        SHOD['halos']['rsr200'] = rs/SHOD['halos']['r200']

    if(SHOD['halo_filetype']=='AbacusRockstar'):
        #initilize the dictionaries with empty values
        for tkey in ['xyz','vxyz','Mvir','sigv','r200','rs','rsr200']:
            SHOD['halos'][tkey]=np.array([])
        Kmap={'xyz':'pos','vxyz':'vel','Mvir':'m_SO', 'sigv':'vrms','r200':'r','rs':'klypin_rs'}
        #Note in Abacus rs and klypin_rs are exactly same for small r but differ for large r
      
        import h5py
        print('*** Loading halos from directory:',SHOD['halo_dir'])
        fnum=0
        hfname='halos_0.%d.h5'%fnum
        print('Warning: not entirely sure about r200 need to check further')
        SHOD['halos']['xyztype']='nonunit' #to say whether halo co-ordinates are in unit box or not
        while(path.isfile(SHOD['halo_dir']+hfname)):
            print(fnum,hfname)
            htmp=h5py.File(SHOD['halo_dir']+hfname)
            for tkey in Kmap.keys():
               SHOD['halos'][tkey]  = accumulate_arrays(SHOD['halos'][tkey],htmp['halos'][Kmap[tkey]])
            htmp.close()
            fnum=fnum+1
            hfname='halos_0.%d.h5'%fnum

        SHOD['halos']['Mhalo']= SHOD['halos']['Mvir']
        SHOD['halos']['rsr200'] = SHOD['halos']['rs']/SHOD['halos']['r200']
        #remove halos with zero velocity dispersion
        ind_sigv_sel=SHOD['halos']['sigv']>0
        print('Removing %d halos with 0 velocity dispersion'%(np.sum(~ind_sigv_sel)))
        ind_Mhalo=SHOD['halos']['Mhalo']>SHOD['Mcut']
        print('Removing %d halos with halos mass below %f'%(np.sum(~ind_Mhalo),SHOD['Mcut']))
        for tkey in SHOD['halos'].keys():
            if(not isinstance(SHOD['halos'][tkey], (np.ndarray))):
                continue #No selection need for non-array
            elif(SHOD['halos'][tkey].shape[0]!=ind_sigv_sel.size):
                continue #Only the array loaded from halo catalog needs selection
            if(len(SHOD['halos'][tkey].shape)==1):
                SHOD['halos'][tkey]= SHOD['halos'][tkey][ind_sigv_sel*ind_Mhalo]
            elif(len(SHOD['halos'][tkey].shape)==2):
                SHOD['halos'][tkey]= SHOD['halos'][tkey][ind_sigv_sel*ind_Mhalo,:]
     

    if(SHOD['halo_filetype']=='Joachimtxt'):
        htmp=load_txt(SHOD['halo_cat'])
    
        #apply any selections if needed
        ind=htmp[:,17]>SHOD['Mcut'] #only positive non-zero halo mass

        #now assign important propwerites
        SHOD['halos']['xyz']  = htmp[ind,:3]*(505.0/3072.0)
        SHOD['halos']['xyztype']='nonunit' #to say whether halo co-ordinates are in unit box or not
        SHOD['halos']['Mvir'] = htmp[ind,17]*(2.8792E+9 / 8.0)
        SHOD['halos']['Mhalo'] = htmp[ind,17]*(2.8792E+9 / 8.0)

    if(SHOD['halo_filetype']=='CITAhalo'):
        from astropy.cosmology import FlatLambdaCDM
        Fhfile           = open(SHOD['halo_cat'],"rb")
        Nhalo            = np.fromfile(Fhfile, dtype=np.int32, count=1)[0]
        RTHMAXin         = np.fromfile(Fhfile, dtype=np.float32, count=1)
        redshiftbox      = np.fromfile(Fhfile, dtype=np.float32, count=1)

        nfloats_perhalo = 10
        npkdata         = nfloats_perhalo*Nhalo
        peakdata        = np.fromfile(Fhfile, dtype=np.float32, count=npkdata)
        peakdata        = np.reshape(peakdata,(Nhalo,nfloats_perhalo))

        #now assign important propwerites
        SHOD['halos']['xyz']  = peakdata[:,:3]
        SHOD['halos']['xyztype']='nonunit' #to say whether halo co-ordinates are in unit box or not
        #SHOD['halos']['vxyz'] = peakdata[:,3:6]

        SHOD['halos']['r200']  = peakdata[:,6]
        Omega_M   = 0.25
        h         = 0.7
        rho       = 2.775e11*Omega_M*h**2

        SHOD['halos']['Mvir']= 4.0/3*np.pi*np.power(SHOD['halos']['r200'],3)*rho
        SHOD['halos']['Mhalo'] = np.copy(SHOD['halos']['Mvir'])

        #convert distance to redshift
        rad_dist=np.sqrt(np.sum(np.power(SHOD['halos']['xyz'],2)))
        #distance redshift relation
        cosmo = FlatLambdaCDM(H0=100*h, Om0=Omega_M)
        zlims=np.linspace(0,5,500)
        dc = cosmo.comoving_distance(zlims).value

        #interpolate distance to redshift
        SHOD['halos']['redshift']=np.interp(rad_dist,dc,zlims)


        #Concetraition is estimated usin Duffy et.al. Table 1
        #https://arxiv.org/pdf/0804.2486.pdf
        concen=7.85 * np.power(SHOD['halos']['Mhalo'] / (2e+12/h),-0.081)
        concen=concen/ np.power(1.0+SHOD['halos']['redshift'],0.71)
        SHOD['halos']['rsr200'] = 1.0/concen


    print('******')
    print('Finished loading halos: nhalos=',SHOD['halos']['xyz'].shape[0])

    #halo mass function
    #mhalo_lims=np.linspace(SHOD['halos']['Mhalo'].min(),SHOD['halos']['Mhalo'].min(),100)
    Nhalo, binhalo = np.histogram(np.log10(SHOD['halos']['Mhalo']),bins=100)
    mhalo_mid=np.power(10,0.5*(binhalo[:-1]+binhalo[1:]))
    SHOD['halos']['halohist']=np.column_stack([mhalo_mid,Nhalo])
    #print('halo mass :', mhalo_mid)

    for props in ['xyz','vxyz']:
       if(props  not in SHOD['halos'].keys()):
          continue

       if('v' in props):
           ccpre='v'
       else:
           ccpre=''

       limstr='halo limits :'
       for ii,cc in enumerate(['x','y','z']):
          limstr='%s %s= %12.5e %12.5e '%(limstr,ccpre+cc,SHOD['halos'][props][:,ii].min(),
                SHOD['halos'][props][:,ii].max())
       print(limstr)
    print('******')

    if('subsamp_halos' in SHOD.keys()): #subsample the halos for efficient runs
        nhalo_orig= SHOD['halos']['xyz'].shape[0]
        nhalo_sub = np.int(SHOD['subsamp_halos']*nhalo_orig)
        selind=np.sort(np.random.choice(nhalo_orig,nhalo_sub,replace=False))

        print('Subsampling halos for efficent caluclations %f'%SHOD['subsamp_halos'])
        print('Number of halos orig= %d sub= %d'%(nhalo_orig,nhalo_sub))
        for tkey in SHOD['halos'].keys():
            if(not isinstance(SHOD['halos'][tkey], (np.ndarray))):
                continue #No selection need for non-array
            elif(tkey in ['halohist']): #things to skip whilw subsampling halos
                continue
            elif(len(SHOD['halos'][tkey].shape)==1):
                SHOD['halos'][tkey]= SHOD['halos'][tkey][selind]
            elif(len(SHOD['halos'][tkey].shape)==2):
                SHOD['halos'][tkey]= SHOD['halos'][tkey][selind,:]


    return SHOD

def accumulate_arrays(arr1,arrin):
    if(arr1.size==0):
        arr1=np.copy(arrin)
    elif(len(arrin.shape)==1):
        arr1=np.concatenate([arr1,arrin])
    elif(len(arrin.shape)==2):
        arr1=np.row_stack([arr1,arrin])

    return arr1
