#Author: Shadab Alam, April 2015
#HOD code with numpy to speed up
#based on my HOD code in c++

import numpy as np
import argparse
from scipy.special import erfc
from scipy import interpolate

import sys
import General_function_HOD as GFH

#Contains functions to deal with specially ELG HOD
import ELG_HOD as ELGHOD

import time
import pylab as pl

#example runs carver:
#python HOD_populate.py -hcat /global/scratch2/sd/shadaba/RunPB/Halo_cat/PB00_0.6452.haloXYZMVSP.txt -redshift 0.55 -vel 1 -LMcut 13.249 -LM1 14.1789 -sigma 0.8975 -kappa 0.136867 -alpha 1.15071 -outroot /global/scratch2/sd/shadaba/RunPB/Gal_cat/PB00_0.6452-Surhudv2M200cfofpy.galXYZFMVP.txt -cfof 1

#example runs coma:
#python HOD_populate.py -hcat /physics/shadaba/Martin_Sim/Halo_cat/PB00_0.6452.haloXYZMVSP-M12.txt  -redshift 0.55 -vel 1 -LMcut 13.249 -LM1 14.1789 -sigma 0.8975 -kappa 0.136867 -alpha 1.15071 -outroot /physics/shadaba/Martin_Sim/Gal_cat/PB00_0.6452-Surhudv2M200cfofpy.galXYZFMVP.txt -cfof 1

if __name__=="__main__":
   #define inputs
   parser = argparse.ArgumentParser(description='Populate halo with galaxies using HOD:')
   #these should be input
   #common parameters
   parser.add_argument('-plots',type=int,default=1)
   parser.add_argument('-sav'  ,type=int,default=1)
   parser.add_argument('-hcat' ,default='/home/salam/Projects/HODFitter/test-bolshoihalos.txt')
   parser.add_argument('-halo_type' ,default='bolshoitxt')
   parser.add_argument('-subhalo' ,type=int,default=1,help='set 0 if subhalo needs to be removed by pid==-1')
   parser.add_argument('-satelite', default='NFW: bolshoiNFW: scalers')
   parser.add_argument('-redshift',type=float, default=0.0)
   parser.add_argument('-xispace'   ,   default='Real', help='set xispace to zRSD to apply RSD along z -axis in the output catalogue')
   parser.add_argument('-OMm'     ,type=float, default=0.292)
   parser.add_argument('-OML'     ,type=float, default=0.708)
   parser.add_argument('-H0'      ,type=float, default=69.0)
   parser.add_argument('-Lbox'    ,type=float, default=250)
   parser.add_argument('-outroot' ,default='testput')
   parser.add_argument('-outgal' ,default='',
           help='''This decides what to write in the output catalog, 
                           "":writes everything , 
                           "basic":writes x,y,z,cen/sat, 
                           "whaloid": To also write halo index ''')

   #This decide how many samples to populate
   parser.add_argument('-samples' ,nargs='+',default=['S1','S2'],
           help='''This is to decide how many samples should be populated ''')

   #sample specific parameter for sample 1
   parser.add_argument('-HODmodel_S1', default='normal',help='To select which HOD model to use normal, ELG, QSO')
   parser.add_argument('-LMcut_S1'   ,type=float, default=13.249)
   parser.add_argument('-LM1_S1'     ,type=float, default=14.1789)
   parser.add_argument('-sigma_S1'   ,type=float, default=0.8975)
   parser.add_argument('-kappa_S1'   ,type=float, default=0.136867)
   parser.add_argument('-alpha_S1'   ,type=float, default=1.15071)
   parser.add_argument('-gamma_S1'   ,type=float, default=1.0,help='skewness in the skewed gaussian for central')
   parser.add_argument('-pmax_S1'   ,type=float, default=1.0,help='set the maximum probability fo central at any halo mass')
   parser.add_argument('-gNFW_S1'   ,type=float, default=1.0,help='gamma of gNFW to get NFW set this to 1.0')
   parser.add_argument('-gammaHV_S1'   ,type=float, default=1.0)
   parser.add_argument('-fconc_S1'   ,type=float, default=1.0)
   parser.add_argument('-gammaIHV_S1'     ,type=float,   default=1.0)
   parser.add_argument('-sigmaCen_S1'     ,type=float,   default=0.0,help='adding error on the central galaxy position in the unit of r200')
   parser.add_argument('-Acen_S1'     ,type=float,   default=0.0,help='Assembly bias with concentration for central')
   parser.add_argument('-Asat_S1'     ,type=float,   default=0.0,help='Assembly bias with concentration for satelite')

   #sample specific parameter for sample 2
   parser.add_argument('-HODmodel_S2', default='ELG',help='To select which HOD model to use normal, ELG')
   parser.add_argument('-LMcut_S2'   ,type=float, default=11.5)
   parser.add_argument('-LM1_S2'     ,type=float, default=13.1789)
   parser.add_argument('-sigma_S2'   ,type=float, default=0.8975)
   parser.add_argument('-kappa_S2'   ,type=float, default=0.136867)
   parser.add_argument('-alpha_S2'   ,type=float, default=1.15071)
   parser.add_argument('-gamma_S2'   ,type=float, default=1.0,help='skewness in the skewed gaussian for central')
   parser.add_argument('-pmax_S2'   ,type=float, default=1.0,help='set the maximum probability fo central at any halo mass')
   parser.add_argument('-gNFW_S2'   ,type=float, default=1.0,help='gamma of gNFW to get NFW set this to 1.0')
   parser.add_argument('-gammaHV_S2'   ,type=float, default=1.0)
   parser.add_argument('-fconc_S2'   ,type=float, default=1.0)
   parser.add_argument('-gammaIHV_S2'     ,type=float,   default=1.0)
   parser.add_argument('-sigmaCen_S2'     ,type=float,   default=0.0,help='adding error on the central galaxy position in the unit of r200')
   parser.add_argument('-Acen_S2'     ,type=float,   default=0.0,help='Assembly bias with concentration for central')
   parser.add_argument('-Asat_S2'     ,type=float,   default=0.0,help='Assembly bias with concentration for satelite')

   #sample specific parameter for sample 3
   parser.add_argument('-HODmodel_S3', default='ELG',help='To select which HOD model to use normal, ELG')
   parser.add_argument('-LMcut_S3'   ,type=float, default=11.5)
   parser.add_argument('-LM1_S3'     ,type=float, default=13.1789)
   parser.add_argument('-sigma_S3'   ,type=float, default=0.8975)
   parser.add_argument('-kappa_S3'   ,type=float, default=0.136867)
   parser.add_argument('-alpha_S3'   ,type=float, default=1.15071)
   parser.add_argument('-gamma_S3'   ,type=float, default=1.0,help='skewness in the skewed gaussian for central')
   parser.add_argument('-pmax_S3'   ,type=float, default=1.0,help='set the maximum probability fo central at any halo mass')
   parser.add_argument('-gNFW_S3'   ,type=float, default=1.0,help='gamma of gNFW to get NFW set this to 1.0')
   parser.add_argument('-gammaHV_S3'   ,type=float, default=1.0)
   parser.add_argument('-fconc_S3'   ,type=float, default=1.0)
   parser.add_argument('-gammaIHV_S3'     ,type=float,   default=1.0)
   parser.add_argument('-sigmaCen_S3'     ,type=float,   default=0.0,help='adding error on the central galaxy position in the unit of r200')
   parser.add_argument('-Acen_S3'     ,type=float,   default=0.0,help='Assembly bias with concentration for central')
   parser.add_argument('-Asat_S3'     ,type=float,   default=0.0,help='Assembly bias with concentration for satelite')

   argsAll = vars(parser.parse_args())

#define functions
def ltime():
   tm=time.localtime()
   tmstr=str(tm.tm_hour)+':'+str(tm.tm_min)+':'+str(tm.tm_sec)
   return tmstr

def Mfof2M200(mfof):
   Lmfof=np.log10(mfof)
   pbest = np.array([1.21128825e+01, 9.33904756e-01, 1.20913807e-02])
   Lm200b = pbest[2]*(Lmfof-12.0)**2 + pbest[1]*(Lmfof-12.0) + pbest[0]
   m200b=np.power(10,Lm200b)
   return m200b

def MeanBCG(args,Mhalo):
   logMcut=np.log(10.0)*args['LMcut']
   Mdiff=(logMcut-np.log(Mhalo))/(np.sqrt(2.0)*args['sigma'])
   NBCG = 0.5 *erfc(Mdiff)*args['pmax']
   return NBCG

def MeanSat(args,Mhalo,NBCG):
   Mdiff=(Mhalo-args['kappa']*np.power(10,args['LMcut']))/np.power(10,args['LM1'])
   Nsat=np.zeros(Mdiff.size)
   index=Mdiff>0
   Nsat[index]=np.power(Mdiff[index],args['alpha'])
   if(args['HODmodel']=='normal'):
       Nsat=Nsat*NBCG
   return Nsat

#interpolation for NFW profile
def NFW_setup(args):
   #copied from Sebastien code which is taken from NFW 1996 paper
   tabm200 =np.array([3.519,20.59,22.01,22.65,26.16,28.15,
              29.67,102.68,224.35, 1109.9,1931.5,3009.7])
   tabr200 =np.array([394,710,726,733,769,788,802,1213,1574,2682,3226,3740])
   tabrsr200=np.array([0.068,0.078,0.088,0.065,0.077,0.124,
              0.131,0.110,0.121,0.151,0.188,0.143])
   tabv200 =np.array([196.0,353.2,361.1,364.6,382.5,392.0,
              398.9,603.4,783.0,1334.,1605.,1861.])
   Logm200  =np.log10(tabm200)
   #Logr200  =np.log10(tabr200)
   #Logrsr200=np.log10(tabrsr200)
   #//critical density with redshift
   rho_c=0.92
   rho_c=rho_c*(args['OMm']*pow(1.0+args['redshift'],3)+args['OML'])
   interp_r200   =interpolate.splrep(Logm200,tabr200, s=0,k=1)
   interp_rsr200 =interpolate.interp1d( Logm200, tabrsr200)

   return interp_r200, interp_rsr200,rho_c

def interp_exterp1d(Mhalo,int_rsr200): #interpolate and linearly extrapolate
   x12=[np.log10(3.524E12)-12.0,np.log10(4.519E12)-12.0]
   x34=[np.log10(3005E12)-12.0,np.log10(3009E12)-12.0]
   y12=int_rsr200(x12)
   y34=int_rsr200(x34)
   slope12= (y12[1]-y12[0])/(x12[1]-x12[0])
   c12=y12[0]-slope12*x12[0]
   slope34= (y34[1]-y34[0])/(x34[1]-x34[0])
   c34=y34[0]-slope34*x34[0]
   index12=np.log10(Mhalo)-12.0>x12[0]
   index34=np.log10(Mhalo)-12.0<x34[1]
   index=index12*index34
   rsr200=np.zeros(Mhalo.size)
   rsr200[index] = int_rsr200(np.log10(Mhalo[index])-12.0)
   xs12=np.log10(Mhalo[np.invert(index12)])-12.0
   rsr200[np.invert(index12)]=slope12*xs12+c12
   xs34=np.log10(Mhalo[np.invert(index34)])-12.0
   rsr200[np.invert(index34)]=slope34*xs34+c34
   if(0):
      pl.plot(np.log10(Mhalo[index])-12.0,rsr200[index],'rs')
      pl.plot(xs12,rsr200[np.invert(index12)],'bo')
      pl.plot(xs34,rsr200[np.invert(index34)],'go')
      pl.show()
   return rsr200

def populate_satelite(args,hcat,hod_dic,hindex,galcat,nbcgsel,rho_c,meanSat,r200,rsr200):
   if(meanSat.size!= r200.size):
      print 'error, number of satelite and r200 doesn\'t match'
      sys.exit(-1)

   nsat=0
   Ngal=nbcgsel

   #get the rho_0 and rs
   if(hcat['satelite']=='NFW'):
      delta_c=hod_dic['fconc']*1.0/rsr200
      delta_c= 200.0*np.power(delta_c,3)/(np.log(1.0+delta_c)-(delta_c/(1.0+delta_c)))
      rho_0=delta_c*rho_c
      rs=r200*rsr200
      r200=hod_dic['fconc']*r200
   elif(hcat['satelite']=='bolshoiNFW'):
      rs=rsr200*r200
      r200=hod_dic['fconc']*r200
      #see Mvir equation in : https://en.wikipedia.org/wiki/Navarro%E2%80%93Frenk%E2%80%93White_profile
      delta_c=hod_dic['fconc']*1.0/rsr200
      delta_c=np.log(1.0+delta_c)-(delta_c/(1.0+delta_c))
      #print(hcat['halos']['Mvir'][hindex].shape,rs.shape)
      rho_0=hcat['halos']['Mvir'][hindex]/(np.pi *4.0* np.power(rs,3))
      rho_0=rho_0/delta_c
   elif(hcat['satelite']=='scalers'):
      rs=rsr200*r200/hod_dic['fconc']
      #see Mvir equation in : https://en.wikipedia.org/wiki/Navarro%E2%80%93Frenk%E2%80%93White_profile
      delta_c=r200/rs
      delta_c=np.log(1.0+delta_c)-(delta_c/(1.0+delta_c))
      #print(hcat['halos']['Mvir'][hindex].shape,rs.shape)
      rho_0=hcat['halos']['Mvir'][hindex]/(np.pi *4.0* np.power(rs,3))
      rho_0=rho_0/delta_c
   else:
      print('Error: Invalid satelite variable')
      sys.exit()

   dr=1.0
   drby2=dr/2.0
   nr=np.floor((r200-1.0)/dr)
   nbin=np.max(nr)+1
   rr=np.linspace(1.0,dr*nbin,nbin)
   dv=4.0*np.pi*(np.power(rr+drby2,3)-np.power(rr-drby2,3))/3.0

   a=1.0/(1+args['redshift']);
   H=args['H0']*np.sqrt(args['OMm']/np.power(a,3) + args['OML'])
   hxyz=hcat['halos']['xyz'][hindex,:]
   hmhalos=hcat['halos']['Mhalo'][hindex]

   #need to track halo index for central/sat connection
   if(args['outgal']=='whaloid'):
      hindex_val=np.where(hindex==1)[0]

   if(galcat.shape[1]>3): # whenever velocity is needed
      #sigV=args['gammaIHV']*hcat['halos']['sigv'][hindex]/(a*H*args['Lbox']) #multiply by velocity bias
      #vhalo=np.sqrt(np.sum(np.power(hcat['halos']['vxyz'][hindex,:],2),axis=1))
      hvxyz=hod_dic['gammaHV']*hcat['halos']['vxyz'][hindex,:]
      hsigv=hod_dic['gammaIHV']*hcat['halos']['sigv'][hindex]
#      hsigv=hcat['halos']['sigv'][hindex]
   for ii in range(0,r200.size):
      #Nsat=np.random.poisson(lam=meanSat[ii])
      Nsat=meanSat[ii]
      #This one is purely NFW
      #rhodv=dv*rho_0[ii]*dr/(rr*np.power(1+(rr/rs[ii]),2)/rs[ii])
      #This one is gNFW where gamma is gNFW and can be set to one to revoer above equation
      rhodv=dv*rho_0[ii]*dr/(np.power(rr/rs[ii],hod_dic['gNFW'])*np.power(1+(rr/rs[ii]),3-hod_dic['gNFW']))
      cumrho=np.cumsum(rhodv[:np.int(nr[ii])],dtype=float)
      if(cumrho[-1]==0):
          continue
      else:
          cumrho=cumrho/cumrho[-1]

      if(0):#to check the satelite probability profiel
         pl.figure() #debug
         print nr[ii],rr[nr[ii]], cumrho.size, rhodv.size
         pl.plot(rr,rhodv,'r') #debug 
         pl.figure()
         pl.plot(rr[:cumrho.size],cumrho,'b') #debug
         pl.show() #debug
      rands=np.random.random(Nsat)
      #This line is a bug replace by the correct line below
      #It was mistakenly taken from older code; I corrected thus on 19 May 2020
      #theta=np.pi*np.random.random(Nsat)
      theta=np.arccos(np.random.random(Nsat)*2-1)
      phi  =2*np.pi*np.random.random(Nsat)
      if(galcat.shape[1]>3): # whenever velocity is needed
         #vsat=np.random.normal(loc=vhalo[ii],scale=sigV[ii],size=Nsat)
         #vfactor=vsat/vhalo[ii]
         vsatx=np.random.normal(loc=hvxyz[ii,0],scale=hsigv[ii],size=Nsat)
         vsaty=np.random.normal(loc=hvxyz[ii,1],scale=hsigv[ii],size=Nsat)
         vsatz=np.random.normal(loc=hvxyz[ii,2],scale=hsigv[ii],size=Nsat)
#         vsatz=args['gammaIHV']*np.random.normal(loc=hvxyz[ii,2],scale=hsigv[ii],size=Nsat)
      for jj in range(0,Nsat): 
         #satelite co-ordinates
         #rand=0.5 #debug
         rsat=rr[np.argmin(np.abs(cumrho-rands[jj]))]/1000.0  #in Mpc
         #print hcat[ii,3], rsat #debug
         if(hcat['halos']['xyztype']=='unit'):
            rsat=rsat/args['Lbox']   #in unit box
         #add entries to galcat if its full in terms of rows
         if(galcat.shape[0]==Ngal):
             if(Ngal>2*hcat['halos']['xyz'].shape[0]):
                 print('****Warning: Too many galaxies!!',ii,Ngal,
                         hcat['halos']['xyz'].shape[0],r200.size,
                         np.sum(meanSat))
             Nextra=np.int(np.sum(meanSat[ii:]))
             #galcat=np.row_stack([galcat,np.zeros(10000*galcat.shape[1]).reshape(10000,galcat.shape[1])])
             galcat=np.row_stack([galcat,np.zeros(Nextra*galcat.shape[1]).reshape(Nextra,galcat.shape[1])])

         galcat[Ngal,0]=hxyz[ii,0]+rsat*np.sin(theta[jj])*np.cos(phi[jj])
         galcat[Ngal,1]=hxyz[ii,1]+rsat*np.sin(theta[jj])*np.sin(phi[jj])
         galcat[Ngal,2]=hxyz[ii,2]+rsat*np.cos(theta[jj])
         #if hcat in unit
         if(hcat['halos']['xyztype']=='unit'):
            galcat[Ngal,:3]=np.mod(galcat[Ngal,:3],1.0)
         else:
            galcat[Ngal,:3]=np.mod(galcat[Ngal,:3],args['Lbox'])

         if(galcat.shape[1]>(3+1)): # whenever velocity is needed
            #satelite velocity
            galcat[Ngal,3]=vsatx[jj] #vfactor[jj]*hvxyz[ii,0]
            galcat[Ngal,4]=vsaty[jj] #vfactor[jj]*hvxyz[ii,1]
            galcat[Ngal,5]=vsatz[jj] #vfactor[jj]*hvxyz[ii,2]

         if(galcat.shape[1]>(6+1)): # extra information is not needed many times
            galcat[Ngal,6]=hsigv[ii]
            galcat[Ngal,7]=hmhalos[ii]#halo mass
            #satelite flag
            galcat[Ngal,8]=10
   
         #need to track halo index for central/sat connection
         if(args['outgal']=='whaloid'):
            galcat[Ngal,args['hidcol']]=hindex_val[ii]

         Ngal=Ngal+1
         nsat=nsat+1
         
         #debug
         #print galcat[Ngal-1,:]
      
   
   return galcat,nsat,Ngal

#Count the expected number of galaxy
def Count_expected_Ngal(args,halohist,boxsize=1):
   if(args['HODmodel'] in ['normal','QSO']):
      meanBCG = MeanBCG(args,halohist[:,0]) #find mean number of BCG based on halo mass
      #meanSat=np.zeros(meanBCG.size)    #initialize mean number of satelite       
      meanSat = MeanSat(args,halohist[:,0],meanBCG) #find mean satelite
   elif(args['HODmodel']=='ELG'):
      #find mean number of BCG based on halo mass
      meanBCG=ELGHOD.CEN_Skew_normal(args,halohist[:,0])
      #find number of satelite
      #meanSat=ELGHOD.Sat_PowerLaw(args,halohist[:,0])
      meanSat=ELGHOD.Sat_PowerLawLMcut(args,halohist[:,0])
   

   #Total number of BCG
   Ntot_BCG = np.sum(meanBCG*halohist[:,1])
   #Total number of satelite
   Ntot_Sat = np.sum(meanSat*halohist[:,1])

   Ngal_theory = Ntot_BCG + Ntot_Sat
   if(0):
      print('Expected number of galaxy BCG,Sat, tot: ',Ntot_BCG,Ntot_Sat,Ngal_theory)
      print('Expected number of galaxy BCG,Sat, tot: ',meanBCG.size,meanSat.size,halohist.shape)
      import pylab as pl
      pl.plot(halohist[:,0],halohist[:,1],'ks-',label='halos')
      pl.plot(halohist[:,0], halohist[:,1]*meanBCG,'r-',label='BCG')
      pl.plot(halohist[:,0], halohist[:,1]*meanSat,'b-',label='Satelite')
      pl.plot(halohist[:,0], halohist[:,1]*(meanBCG+meanSat),'g-',label='BCG+Satelite')
      pl.xscale('log');pl.yscale('log')
      pl.legend(fontsize=14,loc=4)
      pl.show()

   ndensity= Ngal_theory/np.power(boxsize,3)
   return Ngal_theory, ndensity ,meanBCG
#end of count ngal

#define main
def main(args,hcat,hod_S12,samplekeys,galcat='',int_r200='',int_rsr200='',rho_c=0,write=0):
 
   #This is needed for call from emcee MultiHOD
   if('outgal' not in args.keys()):
      args['outgal']=''
      
   nhalo=hcat['halos']['xyz'].shape[0]  #number of halos
   if(write==1):
      print 'Number of halos:', nhalo,ltime()

   #get the meanBCG for S1 and S2
   meanBCG={'tot':np.zeros(hcat['halos']['Mhalo'].size)}; meanSat={}
   for ss, skey in enumerate(samplekeys):
      if(hod_S12[skey]['HODmodel'] in ['normal','QSO']):
         meanBCG[skey] = MeanBCG(hod_S12[skey],hcat['halos']['Mhalo']) #find mean number of BCG based on halo mass
         #meanSat=np.zeros(meanBCG.size)    #initialize mean number of satelite       
         meanSat[skey] = MeanSat(hod_S12[skey],hcat['halos']['Mhalo'],meanBCG[skey]) #find mean satelite
      elif(hod_S12[skey]['HODmodel']=='ELG'):
         #find mean number of BCG based on halo mass
         meanBCG[skey]=ELGHOD.CEN_Skew_normal(hod_S12[skey],hcat['halos']['Mhalo'])
         #find number of satelite
         #meanSat[skey]=ELGHOD.Sat_PowerLaw(hod_S12[skey],hcat['halos']['Mhalo'])
         meanSat[skey]=ELGHOD.Sat_PowerLawLMcut(hod_S12[skey],hcat['halos']['Mhalo'])

      #Adding the assembly bias effects
      if(hod_S12[skey]['Acen']!=0): #Adding assembly bias for central galaxies with concentration
          #based on equation 5 of https://arxiv.org/pdf/1606.07817.pdf
          #median concentration: Note rsr200 is 1/concentration
          med_rsr200=np.median(hcat['halos']['rsr200'])
          dN=hod_S12[skey]['Acen']*np.min(np.column_stack([meanBCG[skey],1-meanBCG[skey]]),axis=1)
          index=hcat['halos']['rsr200']>med_rsr200 #greater this means low concentrration
          #set the low concentration ones dN to - dN
          dN[index]=-dN[index]
          meanBCG[skey]=meanBCG[skey]+dN
      #Do somewhat similar things for satelites
      if(hod_S12[skey]['Asat']!=0): #Adding assembly bias for satelite galaxies with concentration
          #based on equation 5 of https://arxiv.org/pdf/1606.07817.pdf
          #median concentration: Note rsr200 is 1/concentration
          med_rsr200=np.median(hcat['halos']['rsr200'])
          dN=hod_S12[skey]['Asat']*meanSat[skey]
          index=hcat['halos']['rsr200']>med_rsr200 #greater this means low concentrration
          #set the low concentration ones dN to - dN
          dN[index]=-dN[index]
          meanSat[skey]=meanSat[skey]+dN

      meanBCG['tot']=meanBCG['tot']+meanBCG[skey]
   
   #populating BCG
   urand=np.random.random(nhalo)
   index=np.where(urand<meanBCG['tot'])[0]
   nbcgsel=index.size #np.sum(index)


   if(isinstance(galcat,str)):
      ntmp=np.int(1.2*nbcgsel) #expecting 20% satelites
      if(write==1): #number of columns to report
          ncols=9+1
      elif(write==0 and args['xispace'] in ['zRSD']):
          ncols=6+1
      else:
          ncols=3+1

      if('vxyz' not in hcat['halos'].keys()):
          ncols=3+1

      #Note that last column is reserved for galtype so this must go in one before last
      if(args['outgal']=='whaloid'):
          ncols=ncols+1
          args['hidcol']=ncols-2

      galcat=np.zeros(ntmp*ncols).reshape(ntmp,ncols)

   galcat[:nbcgsel,:3]=hcat['halos']['xyz'][index,:]
   #write halo index as id for tracing things back later of satelite central
   if(args['outgal']=='whaloid'):
      galcat[:nbcgsel,args['hidcol']]=index

   #assign galaxy types
   this_rand=urand[index]
   this_BCG=np.zeros(this_rand.size)
   for ss, skey in enumerate(samplekeys):
      ind1=this_rand>this_BCG
      ind2=this_rand<=this_BCG+meanBCG[skey][index]
      ind12=np.where(ind1*ind2==1)[0]
      #assigning galtype here
      galcat[ind12,ncols-1]=ss+1
      this_BCG=this_BCG+meanBCG[skey][index]

      if(galcat.shape[1]>(3+1)): #only redshift space so velocity
         galcat[ind12,3:6]=hod_S12[skey]['gammaHV']*hcat['halos']['vxyz'][index[ind12],:]
         #print('gpar:',args['gammaHV'],args['gammaIHV'])

      if(galcat.shape[1]>(6+1)): #need all info to write catalog
         galcat[ind12,6]=hod_S12[skey]['gammaIHV']*hcat['halos']['sigv'][index[ind12]]
         galcat[ind12,7]=hcat['halos']['Mhalo'][index[ind12]]
         galcat[ind12,8]=100


      #Adding error in the central galaxy position as per sigmaCen values in unit of r200
      if(hod_S12[skey]['sigmaCen']!=0):
          err=hod_S12[skey]['sigmaCen']*hcat['halos']['r200'][index[ind12]]
          #assuming r200 is in kpc/h so dividing it by 1000 to get it in Mpc/h same as xyz
          err=err/1000.0 #This 1 sigma of the r err with gaussian distribution
          #generate random number for radius
          err=np.abs(np.random.normal(0,err))
          #generate random phi
          phi_err=np.random.random(err.size)*np.pi*2.0
          theta_err=np.arccos(np.random.random(err.size)*2.0 -1)

          #update central galaxy position with additional error
          galcat[ind12,0]=np.mod(galcat[ind12,0]+(err*np.sin(theta_err)*np.cos(phi_err)),args['Lbox'])
          galcat[ind12,1]=np.mod(galcat[ind12,1]+(err*np.sin(theta_err)*np.sin(phi_err)),args['Lbox'])
          galcat[ind12,2]=np.mod(galcat[ind12,2]+(err*np.cos(theta_err)),args['Lbox'])


   if(write==1):
     print 'Number of BCG:' ,nbcgsel,ltime()

   Ngal_cur=nbcgsel
   Ngal=nbcgsel
   #populating satelite
   for ss, skey in enumerate(samplekeys):
      Nsat=np.random.poisson(lam=meanSat[skey])
      index=Nsat>0   #halos with non-zero satelites

      if(np.sum(index)<2):
         continue
         #return galcat,nbcgsel,int_r200,int_rsr200,rho_c


      #Figure out rs and rsr200 for NFW
      if('r200' not in hcat['halos'].keys() or 
           'rsr200' not in hcat['halos'].keys()):
          if(int_r200=='' or int_rsr200=='' or rho_c==0): #interpolation needed
             int_r200,int_rsr200,rho_c=NFW_setup(args)
             print('*** Interpolating the r200 and rsr200 for NFW')
   
          r200=interpolate.splev(np.log10(hcat['halos']['Mhalo'][index])-12.0,int_r200, der=0)
          #rsr200 = interpolate.int_rsr200(np.log10(hcat[index,Mind])-12.0)
          rsr200 = interp_exterp1d(hcat['halos']['Mhalo'][index],int_rsr200)  
          #interpolate and linearly extrapolate
          galcat,nsatsel,Ngal=populate_satelite(args,hcat,hod_S12[skey],index,galcat,
             Ngal_cur,rho_c,Nsat[index],r200,rsr200)
      else:
          galcat,nsatsel,Ngal=populate_satelite(args,hcat,hod_S12[skey],index,galcat,
             Ngal_cur,0,Nsat[index],
             hcat['halos']['r200'][index],hcat['halos']['rsr200'][index])

      #assign the galaxy type for satelites
      galcat[Ngal_cur:Ngal,-1]=ss+1
      #update Ngal_cur
      Ngal_cur=Ngal

   #convert co-ordinates to redshift space if needed
   if(args['xispace']=='zRSD'):
       Hz=100.0 # so that we get result in Mpc/h, velocities assumed in km/s
       galcat[:,2]=galcat[:,2]+(galcat[:,5]*(1+args['redshift'])/Hz)
       #apply modulo box size
       galcat[:,2]=np.mod(galcat[:,2],args['Lbox'])

   if(write==1):
      print 'Number of Sat:' ,nsatsel,ltime()
      header='Galaxy Catalogue:\nInputs:\n'
      names=args.keys()
      stage=50
      nstage=1
      for nn in names:
         this_nn=0
         if(nn[0]=='_'):
            continue
         elif('_S' in nn):
            for ss,skey in enumerate(samplekeys):
               if(skey in nn):
                  this_nn=0; break
               else:
                  this_nn=1
            if(this_nn==1):
               continue

         if(this_nn==0):
            val=args[nn]
            if(type(val)!=type('a')):
               val=str(val)
            header=header+nn+'='+val+' '
            if(len(header)>nstage*stage):
               header=header+'\n'
               nstage=nstage+1

      if(args['outgal'] in ['','whaloid']):
         if(args['outgal']==''):
            header=header+"\n#columns are: x,y,z,Vx,Vy,Vz,sigV,Mhalo,flag, galtype"
         elif(args['outgal']=='whaloid'):
            header=header+"\n#columns are: x,y,z,Vx,Vy,Vz,sigV,Mhalo,flag, haloindex, galtype"

         header=header+"\n#flag=10 satelite galaxy, flag=100 BCG galaxy\n"
         fmts=''
         for ii in range(0,galcat.shape[1]):
             if(args['outgal']=='whaloid' and args['hidcol']==ii):
                fmts=fmts+'%15d '
             else:
                fmts=fmts+'%16.9g '

         np.savetxt(hcat['outroot']+'.gcat',galcat[:Ngal,:],header=header,fmt=fmts)
      elif(args['outgal']=='basic'):
         header=header+"\n#columns are: x,y,z,Vx,Vy,Vz,cen/sat"
         #if(args['vel']==1):
         header=header+"\n#cen/sat=10 satelite galaxy, cen/sat=100 central galaxy\n"
         fmts=''
         for ii in range(0,6):
             fmts=fmts+'%16.8g '
         fmts=fmts+'%4d '
         np.savetxt(hcat['outroot']+'.gcat',galcat[:Ngal,[0,1,2,3,4,5,8]],header=header,fmt=fmts)

      print 'gal catalogue:',hcat['outroot']+'.gcat',ltime()

  

   return galcat,Ngal,int_r200,int_rsr200,rho_c

#call main
#load catalogue
if __name__=="__main__":
   #dictionary for halo catalog
   print 'loading halo catalogue:', argsAll['hcat'],ltime()
   #hcat=np.loadtxt(argsAll['hcat'])
   hcat={'halo_filetype':argsAll['halo_type'],'halo_cat':argsAll['hcat'],
           'Lbox':argsAll['Lbox'],'satelite':argsAll['satelite'],'Mcut':1e9,
           'outroot':argsAll['outroot'],'subhalo':argsAll['subhalo']}

   samplekeys=argsAll['samples']
   #initiate the HOd parameter dictionaries
   hod_S={}
   for ss,skey in enumerate(samplekeys):
      hod_S[skey]={}

   #split the hod parameters in two dictionary for two samples
   for kk, tkey in enumerate(argsAll.keys()):
      skey=tkey.split('_')[-1]
      if(skey in samplekeys):
         hod_S[skey][tkey[:-(len(skey)+1)]]=argsAll[tkey]


   for ss,skey in enumerate(samplekeys):
      print('hod_%s'%skey,hod_S[skey])

   hcat=GFH.load_halos(hcat)
   print 'finished loading halo catalogue:',ltime()

   #Check for consistency between Lbox and halo catalogue
   blen=np.max(hcat['halos']['xyz'])
   if(blen-argsAll['Lbox']>1e-3):
      print('The Lbox and the maximum scale for halo catalog do not match')
      print('Lbox= ',argsAll['Lbox'])
      print('max halo xyz=',blen)
      sys.exit()

   #galcat=np.zeros(1000*9).reshape(1000,9) #debug
   galcat,Ngal,int_r200,int_rsr200,rho_c=main(argsAll,hcat,hod_S,samplekeys,galcat='',int_r200='',int_rsr200='',rho_c=0,write=1)
   print 'finished populating:',ltime()

   #compute histograms
   galcat=galcat[:Ngal,:]
   indexBCG=galcat[:,8]==100
   indexSat=galcat[:,8]==10
   print 'number of bcg,sat,ngal=',np.sum(indexBCG),np.sum(indexSat),Ngal

   Mhalo_bin=np.linspace(10.5,16.5,100)
   histHalo,Mhalo_jnk  =np.histogram(np.log10(hcat['halos']['Mhalo']),bins=Mhalo_bin)

   BCGhist,Mhalo_jnk=np.histogram(np.log10(galcat[indexBCG,7]),bins=Mhalo_bin)
   Sathist,Mhalo_jnk=np.histogram(np.log10(galcat[indexSat,7]),bins=Mhalo_bin)

   LMcentre=0.5*(Mhalo_bin[:-1]+Mhalo_bin[1:])
   Theory_header=''
   hist_header=''
   for ss, skey in enumerate(samplekeys):
      #Histogram of numerical galaxies
      ind_this=galcat[:,-1]==ss+1
      BCGhist,Mhalo_jnk=np.histogram(np.log10(galcat[indexBCG*ind_this,7]),bins=Mhalo_bin)
      Sathist,Mhalo_jnk=np.histogram(np.log10(galcat[indexSat*ind_this,7]),bins=Mhalo_bin)
      print('***** sample %s: \nnumber of bcg= %d ,sat= %d ,ngal=%d'%(skey,np.sum(indexBCG*ind_this),
                   np.sum(indexSat*ind_this),np.sum(ind_this)))

      #Theoretical occupation
      if(hod_S[skey]['HODmodel'] in ['normal','QSO']):
         TBCG_this=MeanBCG(hod_S[skey],np.power(10,LMcentre))
         TSat_this=MeanSat(hod_S[skey],np.power(10,LMcentre),TBCG_this)
      elif(hod_S[skey]['HODmodel']=='ELG'):
         #find mean number of BCG based on halo mass
         TBCG_this=ELGHOD.CEN_Skew_normal(hod_S[skey],np.power(10,LMcentre))
         #find number of satelite
         #TSat_this=ELGHOD.Sat_PowerLaw(hod_S[skey],np.power(10,LMcentre))
         TSat_this=ELGHOD.Sat_PowerLawLMcut(hod_S[skey],np.power(10,LMcentre))

      Theory_header='%s  TheoryBCG(%s) ,   TheorySat(%s), '%(Theory_header,skey,skey)
      hist_header='%s  NBCG(%s),    NSat(%s), '%(hist_header,skey,skey)
      if(ss==0):
          Tgal=np.column_stack([TBCG_this,TSat_this])
          Hgal=np.column_stack([BCGhist,Sathist])
      else:
          Tgal=np.column_stack([Tgal,TBCG_this,TSat_this])
          Hgal=np.column_stack([Hgal,BCGhist,Sathist])

   #A header of parameters
   parheader=''
   for ss, skey in enumerate(np.sort(hod_S.keys())):
      parheader=parheader+'#HOD model sample %s:\n#'%(skey)
      tcount=0
      for kk, tkey in enumerate(np.sort(hod_S[skey].keys())):
         #print(skey,tkey,isinstance(hod_S12[skey][tkey],str))
         parheader='%s %s= '%(parheader,tkey)+str(hod_S[skey][tkey])
         if((kk+1)%3==0):
            parheader=parheader+'\n#'

   #write histgoram
   fname=argsAll['outroot']+'.hist'
   print 'writing histogram:' ,fname
   fhist=open(fname,'w')
   fhist.write("#Statistics for the HOD galaxy catalogue z="+str(argsAll['redshift'])+'\n')
   fhist.write(parheader+'\n')
   fhist.write("#columns: log10(Mhalo),  Nhalo,  %s  %s\n"%(hist_header,Theory_header))
   for ii in range(0,LMcentre.size):
      fhist.write("%12.5f  %10d " %(LMcentre[ii], histHalo[ii]))#, BCGhist[ii], Sathist[ii]))
      for jj in range(0,Hgal.shape[1]):
         fhist.write("%10d "%(Hgal[ii,jj]))
      for jj in range(0,Tgal.shape[1]):
         fhist.write("%12.5f "%(Tgal[ii,jj]))
      fhist.write("\n")

   fhist.close()
   
   #This is to make plot of distributions
   if(1):
      pl.plot(LMcentre,histHalo,'k-',label='halo')
      for jj in range(0,Hgal.shape[1]):
         pl.plot(LMcentre,Hgal[:,jj],label='H %d'%jj) 
         pl.plot(LMcentre,Tgal[:,jj]*histHalo,label='T %d'%jj)
      pl.yscale('log') 
      pl.legend()
      fname=argsAll['outroot']+'.hist.png'
      pl.savefig(fname)
      print('******* Waiting for plot to close: Finished *********')
      pl.show()




