#author: Shadab Alam, convert halo catalog to fits file
from __future__ import print_function

import os
import numpy as np
import pylab as pl

import fitsio as F
try:
    import pandas as pd
    pd_flag=1
except:
    pd_flag=0
    print('''pandas not found: you can install pandas from
            https://pypi.python.org/pypi/pandas/0.18.0/#downloads
            The code will work without pandas but pandas will speed up loading large text files
            ''')

#convert to fits in organized way
import time
import argparse

def convert_fits(source_file,outfile,nlim=100000,sim='MDPL2'):
    
    #This scans the Rocktar halos file and write to fits
    
    Fdic_out={}
    
    ts1=time.localtime()
    print('working on input: ',source_file,time.asctime(ts1))
    print('output fits file: ',outfile)

    #open the text file with pandas iterator
    tb=pd.read_table(source_file,delim_whitespace=True,comment='#',header=None,
          dtype=np.float,chunksize=nlim)
    count=0
    for chunk in tb:
        if(count%5000==0):
            ts2=time.localtime()
            print(count,time.asctime(ts2))
        count=count+1
        
        #Fdic_out=Write2FITS(Fdic_out,datain=chunk.as_matrix(),outfile=outfile,sim=sim)
        Fdic_out=Write2FITS(Fdic_out,datain=chunk.values,outfile=outfile,sim=sim)
        #if(count>10):
        #    break

    #break
    #cleanup by closing the fits file and writing checksum
    ts2=time.localtime()
    print('**** closed: ',Fdic_out['fname'],time.asctime(ts2))
    Fdic_out['fhandle'][-1].write_checksum()
    Fdic_out['fhandle'].close() 

    return 0

def Write2FITS(dicIN,datain=np.array([]),outfile='',sim='MDPL2'):
    '''This takes the input data and write the relevant properties to FITS file'''
    
    #map needed for lightcone angle selection
    lcAng_map={'xx0':3,'xx1':4,'xx2':5}
    
    #map to get the columns for fields
    if(sim=='MDPL2'):
        colTokey={'xx':17,'yy':18,'zz':19,'vx':20,'vy':21,'vz':22,
                  'Mvir':10,'vrms':13,'rvir':11,'rs':12,'Rs_Klypin':37,
                 'M200b':39, 'M200c':40,'M2500c':42,'M500c':41,
                 'pid':5,'upid':6}
    elif(sim=='Bolshoi'):
        colTokey={'xx':17,'yy':18,'zz':19,'vx':20,'vy':21,'vz':22,
                  'Mvir':10,'vrms':13,'rvir':11,'rs':12,'Rs_Klypin':35,
                 'M200b':37, 'M200c':38,'M2500c':40,'M500c':39,
                 'pid':5,'upid':6}
    
    #datatye to write to fits
    outtype=[]
    for kk,tkey in enumerate(colTokey.keys()):
        outtype.append((tkey,'>f4'))
    #print(outtype)    
    
    rowsize=datain.shape[0]
    if('rowsize' not in dicIN):
        dicIN['rowsize']=np.int(datain.shape[0]*2)
        dicIN['dataout']=np.zeros(dicIN['rowsize'],dtype=outtype)

        
    if(rowsize>dicIN['rowsize']):
        dicIN['rowsize']=rowsize
        dicIN['dataout']=None
        dicIN['dataout']=np.zeros(rowsize,dtype=outtype)


    #transfer datain to data out
    for kk in colTokey.keys():
        dicIN['dataout'][kk][:rowsize]=datain[:,colTokey[kk]]

    #If fits file haven't been opened yet the open the fits file
    
    #dictionary to store file handles
    if('fhandle' not in dicIN.keys()):
        dicIN['fname']=outfile
        dicIN['fhandle']=F.FITS(outfile,'rw')
        dicIN['fhandle'].write(dicIN['dataout'][:rowsize])
    else:
        dicIN['fhandle'][-1].append(dicIN['dataout'][:rowsize])

    return dicIN



if __name__=="__main__":
    parser = argparse.ArgumentParser(description='Converts Halo catalog to fits file')
    #these should be input
    parser.add_argument('-halo_cat',default='')
    parser.add_argument('-output_fits',default='')
    args = parser.parse_args()

    #check inputs
    if(args.halo_cat==''):
        raise ValueError("No input halo catalog is provided\n please use -halo_cat option to provide input halo catalog")
    elif(not os.path.isfile(args.halo_cat)): 
        raise ValueError('input halo catalog:%s \n Not Found\n'%args.halo_cat)

    if(args.output_fits==''):
        print('Warning:no output file provided')
        args.output_fits=args.halo_cat+'.fits'
    elif(os.path.isfile(args.output_fits)):
        raise ValueError('output halo catalog:%s \n already exists\n Please provide a different name or remove the existing output file\n\n'%args.output_fits)
    elif(args.output_fits[-5:]!='.fits'):
        raise ValueError('output_fits file name must end with .fits\n\n Please provide appropriate file name')


    sim='MDPL2'
    convert_fits(args.halo_cat, args.output_fits, nlim=100000,sim=sim)
    
    print('Finished!')

